/*
 * klm.c	Handle KLM requests and forward them to the appropriate
 *		NLM server.
 *
 *		This code is not very robust yet. It should check for
 *		conflicting requests from the kernel (i.e. concurrent
 *		LOCKs and UNLOCKs for the same file and region).
 *
 * Copyright (C) 1995, O. Kirch <okir@monad.swb.de>
 */
#include <rpc/rpc.h>
#include <arpa/inet.h>
#include <signal.h>
#include "klm.h"
#include "nlm.h"

struct timeval	clnt_grace_backoff = { 5, 0 };

static int	klm_test_status(int status, lockd_req *rq);
static int	klm_lock_status(int status, lockd_req *rq);
static int	klm_unlock_status(int status, lockd_req *rq);
static int	klm_cancel_status(int status, lockd_req *rq);

/*
 * NULL procedure.
 */
bool_t
klm_null_svc(struct svc_req *rqstp, void *argp, void *resp)
{
	return 1;
}

/*
 * Test for the presence of a lock. Maybe we should check if the target
 * host refers to ourself and take a short cut.
 */
bool_t
klm_test_svc(struct svc_req *rqstp, klm_testargs *argp, klm_testrply *resp)
{
	lockd_host	*host;
	klm_lock *	alock;
	int		ex;

	if ((host = host_lookup(NULL, argp->alock.server_name, 0)) == NULL) {
		resp->stat = klm_granted;
		return 1;
	}

	alock = &(argp->alock);
	ex = argp->exclusive;

	/* For duplicate requests, ignore the call. */
	if (req_find(host->l_waiting, RQ_TEST, alock, ex, 0)) {
		return 0;
	}

	/*
	 * Send the request and insert it into host->l_waiting. If RPC 
	 * cannont send the request, the callback is invoked which
	 * returns an error to the kernel.
	 */
	req_queue(host, rqstp, klm_test_status, RQ_TEST, alock, ex, 0);

	return 0;
}

/*
 * Status callback for a request we sent above. 
 * the call to the server was successful; RPC_CALLBACK says the server's
 * result callback has been received.
 *
 * If the call failed, we propagate the error code to the client. 
 */
static int
klm_test_status(int status, lockd_req *rq)
{
	if (status == RQ_CALLBACK) {
		klm_testrply	reply;
		klm_holder *	kholder;
		nlm_holder *	nholder;
		nlm_testres *	resp;

		resp = (nlm_testres *) (rq->result);
		log(L_DEBUG, "klm received test reply status %d", resp->stat);

		switch (resp->stat.stat) {
		case LCK_GRANTED:
			reply.stat = klm_granted;
			break;
		case LCK_DENIED:
			kholder = &(reply.klm_testrply_u.holder);
			nholder = &(resp->stat.nlm_testrply_u.holder);
			reply.stat = klm_denied;
			kholder->exclusive = nholder->exclusive;
			kholder->svid = nholder->svid;
			kholder->l_offset = nholder->l_offset;
			kholder->l_len = nholder->l_len;
			break;
		case LCK_DENIED_GRACE_PERIOD:
			reply.stat = klm_working;
			break;
		default:
			reply.stat = klm_denied_nolocks;
			break;
		}

		rpcmsg_sendreply(&(rq->caller), rq->sockfd, rq->xid,
					(xdrproc_t)xdr_klm_testrply,
					(caddr_t)&reply);
	} else {
		rpcmsg_failreply(&(rq->caller), rq->sockfd, rq->xid,
					RPC_FAILED);
	}
	return 1;
}

/*
 * Create a lock on the server host.
 * If it succeeds, we must record this lock so we can reclaim it in case
 * of a server crash. Not implemented yet.
 */
bool_t
klm_lock_svc(struct svc_req *rqstp, klm_lockargs *argp, klm_stat *resp)
{
	lockd_host *	host;
	klm_lock *	alock = &(argp->alock);
	int		ex = argp->exclusive;
	int		bl = argp->block;

	if ((host = host_lookup(rqstp, argp->alock.server_name, 0)) == NULL) {
		svcerr_systemerr(rqstp->rq_xprt);
		return 0;
	}

	/* Ignore duplicate requests. */
	if (req_find(host->l_waiting, RQ_LOCK, alock, ex, bl) ||
	    req_find(host->l_reclaim, RQ_LOCK, alock, ex, bl) ||
	    req_find(host->l_blocked, RQ_LOCK, alock, ex, bl)) {
		return 0;
	}

	/* Forward the request to the NLM server. */
	req_queue(host, rqstp, klm_lock_status, RQ_LOCK, alock, ex, bl);

	return 0;
}

/* 
 * Status callback for lock request.
 */
static int
klm_lock_status(int status, lockd_req *rq)
{
	int	ret = 1;

	if (status == RQ_CALLBACK) {
		lockd_host *	host = rq->host;
		nlm_res *	resp;
		klm_stat	reply;

		resp = (nlm_res *) rq->result;

		if (rq->reclaim) {
			if (resp->stat.stat == LCK_GRANTED) {
				/* Successfully reclaimed lock */
				return 0;
			}

			/* Bad things happening. Complain and send lock
			 * holder a SIGLOST.
			 */
			log(L_WARNING, "failed to reclaim lock, status %d",
					resp->stat.stat);
			kill(SIGLOST, rq->pid);
			return 1;
		}

		switch (resp->stat.stat) {
		case LCK_GRANTED:
			host_monitor(host);
			reply.stat = klm_granted;
			ret = 0;
			break;
		case LCK_DENIED:
			reply.stat = klm_denied;
			break;
		case LCK_BLOCKED:
			req_insert(&(host->l_blocked), rq);
			reply.stat = klm_working;
			break;
		case LCK_DENIED_GRACE_PERIOD:
			rq->wait = clnt_grace_backoff;
			req_retry(rq);
			return 0; /* don't reply to client, retry later */
		default:
			reply.stat = klm_denied;
			break;
		}

		rpcmsg_sendreply(&(rq->caller), rq->sockfd, rq->xid,
					(xdrproc_t)xdr_klm_stat,
					(caddr_t)&reply);
	} else {
		rpcmsg_failreply(&(rq->caller), rq->sockfd, rq->xid, 
					RPC_FAILED);
	}
	return ret;
}

/*
 * Cancel a blocked lock on the server host.
 */
bool_t
klm_cancel_svc(struct svc_req *rqstp, klm_lockargs *argp, klm_stat *resp)
{
	lockd_host	*host;
	lockd_req *	rq;
	klm_lock *	alock = &(argp->alock);
	int		ex = argp->exclusive;
	int		bl = argp->block;

	if ((host = host_lookup(rqstp, argp->alock.server_name, 0)) == NULL) {
		svcerr_systemerr(rqstp->rq_xprt);
		return 0;
	}

	/* Check for request retransmissions */
	if (req_find(host->l_waiting, RQ_CANCEL, alock, ex, bl) ||
	    req_find(host->l_waiting, RQ_UNLOCK, alock, ex, bl)) {
		return 0;
	}

	/* This part here is a little tricky. The following races can
	 * occur:
	 *  a.	The lock request is still in flight. Bad things happen
	 *	if the cancel request arrives at the server before the
	 *	lock request.
	 *
	 *  b.	The lock request has been granted already.
	 *
	 * We solve a) by ignoring it (for the moment). To solve b)
	 * we should check if we currently hold this lock and send out
	 * an UNLOCK if we do. However, for the time being, it's easier
	 * to just send the UNLOCK no matter what.
	 */

	/* a: discard any pending LOCK requests. Not optimal. */
	if ((rq = req_find(host->l_waiting, RQ_LOCK, alock, ex, bl)) != NULL) {
		req_unlink(rq);
		req_abort(rq);
		req_free(rq);
	}

	/* b: send an UNLOCK request */
	req_queue(host, rqstp, klm_cancel_status, RQ_UNLOCK, alock, ex, bl);

	/* Check if there's a blocked lock request at all. If there is none,
	 * it may be that we already processed the cancel, but our reply to
	 * the client got lost. Just tell it that all is well.
	 */
	if ((rq = req_find(host->l_blocked, RQ_LOCK, alock, ex, bl)) == NULL) {
		resp->stat = klm_granted;
		return 1;
	}

	req_unlink(rq);
	req_free(rq);

	/* Forward the request to the NLM server. */
	req_queue(host, rqstp, klm_cancel_status, RQ_CANCEL, alock, ex, bl);

	return 0;
}

/*
 * RPC status for an asynchronous NLM_CANCEL_MSG call to the server.
 */
static int
klm_cancel_status(int status, lockd_req *rq)
{
	if (status == RQ_CALLBACK) {
		lockd_host *	host = rq->host;
		klm_stat	reply;
		nlm_res	*	resp;

		resp = (nlm_res *) rq->result;
		switch (resp->stat.stat) {
		case LCK_GRANTED:
			host_unmonitor(host);
			reply.stat = klm_granted;
			break;
		case LCK_DENIED:
			reply.stat = klm_denied;
			break;
		case LCK_DENIED_GRACE_PERIOD:
			rq->wait = clnt_grace_backoff;
			req_retry(rq);
			return 0;
		default:
			reply.stat = klm_denied;
			break;
		}

		rpcmsg_sendreply(&(rq->caller), rq->sockfd, rq->xid,
					(xdrproc_t)xdr_klm_stat,
					(caddr_t)&reply);
	} else {
		rpcmsg_failreply(&(rq->caller), rq->sockfd, rq->xid, 
					RPC_FAILED);
	}
	return 1;
}

/*
 * Free a lock on the server host.
 */
bool_t
klm_unlock_svc(struct svc_req *rqstp, klm_unlockargs *argp, klm_stat *resp)
{
	lockd_host	*host;
	lockd_req *	rq;
	klm_lock *	alock = &(argp->alock);

	if ((host = host_lookup(rqstp, argp->alock.server_name, 0)) == NULL) {
		svcerr_systemerr(rqstp->rq_xprt);
		return 0;
	}

	/*
	 * If there's a matching blocked request, just remove it. 
	 */
	if ((rq = req_find(host->l_blocked, RQ_LOCK, alock, 1, -1))) {
		log(L_WARNING, "weird: unlock request for blocked lock");
		req_unlink(rq);
		req_free(rq);

		resp->stat = klm_granted;
		return 1;
	}

	/* Don't send out duplicate requests. */
	if (req_find(host->l_waiting, RQ_UNLOCK, alock, -1, -1)) {
		log(L_DEBUG, "ignored duplicate KLM_UNLOCK request");
		return 0;
	}

	/* Forward the request to the NLM server. */
	req_queue(host, rqstp, klm_unlock_status, RQ_UNLOCK, alock, 0, 0);

	return 0;
}

/*
 * Status of an NLM_UNLOCK_MSG call to the server.
 */
static int
klm_unlock_status(int status, lockd_req *rq)
{
	if (status == RQ_CALLBACK) {
		lockd_host *	host = rq->host;
		klm_stat	reply;
		nlm_res	*	resp;

		resp = (nlm_res *) rq->result;
		switch (resp->stat.stat) {
		case LCK_GRANTED:
			/* FIXME: remove local matching lock */
			host_unmonitor(host);
			reply.stat = klm_granted;
			break;
		case LCK_DENIED:
			reply.stat = klm_denied;
			break;
		case LCK_DENIED_GRACE_PERIOD:
			rq->wait = clnt_grace_backoff;
			req_retry(rq);
			return 0;
		default:
			reply.stat = klm_denied;
			break;
		}

		rpcmsg_sendreply(&(rq->caller), rq->sockfd, rq->xid,
					(xdrproc_t)xdr_klm_stat,
					(caddr_t)&reply);
	} else {
		rpcmsg_failreply(&(rq->caller), rq->sockfd, rq->xid, 
					RPC_FAILED);
	}
	return 1;
}
