/*
 * nlm_client.c		These functions implement the client side of the
 *			NLM server.
 *
 * Requests to be sent via NLM are wrapped up in lockd_req structs. The
 * bookkeeping for these structs is done almost entirely by klm.c and
 * request.c. The reason to separate this is that I want to be able to
 * support other locking protocols, too. (Don't know what SMB does, but it
 * should have something like this. Maybe DOS file sharing or some such).
 *
 *
 * Copyright (C) 1995, Olaf Kirch  <okir@monad.swb.de>
 */

#include <arpa/inet.h>
#include "nlm.h"

/*
static int	nlm_process_result(struct svc_req *rqstp, void * argp);
static void	nlm_update_local(int action, struct lockd_req *rq);
static int	nlm_grant_blocked(lockd_host *host, nlm_testargs * argp);
 */
static void	nlm_create_oh(netobj *obj, int pid);
static int	nlm_rqinit_1(lockd_req *rq);
static int	nlm_rqsend_1(lockd_req *rq);
static int	nlm_getconflock_1(lockd_req *rq, struct lklm_holder *hop);
/* static int	nlm_rqstatus(int nlmstat); */

lockd_proto	proto_nlm1 = {
	nlm_rqinit_1, nlm_rqsend_1, nlm_getconflock_1
};

#if 0
/*
 * The server returns the result of a TEST_MSG call.
 */
bool_t
nlm_test_res_svc (struct svc_req * rqstp, nlm_testres * argp, void * resp)
{
	return nlm_process_result(rqstp, argp);
}

/*
 * Server returns the result of a LOCK_MSG call.
 * Before returning the result to the original caller, we have to add
 * the lock to our own list of lock requests so that we're able to do
 * lock recovery.
 */
bool_t
nlm_lock_res_svc (struct svc_req * rqstp, nlm_res * argp, void * resp)
{
	return nlm_process_result(rqstp, argp);
}

/*
 * Server returns the result of a CANCEL_MSG call.
 */
bool_t
nlm_cancel_res_svc (struct svc_req * rqstp, nlm_res * argp, void * resp)
{
	return nlm_process_result(rqstp, argp);
}

/*
 * Server returns the result of an UNLOCK_MSG call.
 */
bool_t
nlm_unlock_res_svc (struct svc_req * rqstp, nlm_res * argp, void * resp)
{
	return nlm_process_result(rqstp, argp);
}

static int
nlm_process_result(struct svc_req *rqstp, void * argp)
{
	nlm_res *	resp = (nlm_res *) argp;
	lockd_req *	rq;
	int		status, cookie;

	if (!netobj_get(&(resp->cookie), &cookie, sizeof(cookie))) {
		svcerr_decode(rqstp->rq_xprt);
		return 0;
	}

	if ((rq = req_match(cookie)) == NULL) {
		log(L_DEBUG, "received NLM result for unknown cookie %d",
						cookie);
		return 1;
	}

	status = resp->stat.stat;
	rq->result = argp;

	if (rq->type == RQ_LOCK && status == LCK_GRANTED)
		nlm_update_local(AC_LOCK, rq);
	else if (rq->type == RQ_UNLOCK && status == LCK_GRANTED)
		nlm_update_local(AC_UNLOCK, rq);

	req_process_result(rq, nlm_rqstatus(status));

	return 1;
}

/*
 * The server grants a previously blocked lock.
 */
bool_t
nlm_granted_svc (struct svc_req *rqstp, nlm_testargs * argp, nlm_res * resp)
{
	lockd_host *	host;

	if ((host = host_lookup(rqstp, argp->alock.caller_name, 1)) == NULL) {
		svcerr_systemerr(rqstp->rq_xprt);
		return 0;
	}

	/*
	 * Try to grant the lock request.
	 * If there's no blocked lock to grant, return LCK_DENIED.
	 *
	 * FIXME: Maybe we should follow up with an UNLOCK call instead.
	 */
	if (nlm_grant_blocked(host, argp)) {
		resp->stat.stat = LCK_GRANTED;
	} else {
		resp->stat.stat = LCK_DENIED;
	}
	netobj_init(&(resp->cookie), &(argp->cookie));

	return 1;
}

/*
 * Server grants a previously blocked lock (async version).
 */
bool_t
nlm_granted_msg_svc(struct svc_req * rqstp, nlm_testargs * argp, void * nresp)
{
	lockd_host *	host;
	nlm_res		res;
	int		status;

	if ((host = host_lookup(rqstp, argp->alock.caller_name, 1)) == NULL) {
		svcerr_systemerr(rqstp->rq_xprt);
		return 0;
	}

	/* See the comments in nlm_granted_svc above */
	if (nlm_grant_blocked(host, argp, &res)) {
		res.stat.stat = LCK_GRANTED;
	} else {
		res.stat.stat = LCK_DENIED;
	}
	netobj_init(&res.cookie, &argp->cookie);

	status = rpcmsg_send(host->addr, nlm_svcpg, NLM_VERS, NLM_GRANTED_RES,
				(xdrproc_t) xdr_nlm_res, &res);

	return 1;
}

/* 
 * The server has granted a previously blocked request. 
 */
static int
nlm_grant_blocked(lockd_host *host, nlm_testargs * argp)
{
        struct lklm_lock        klock, *alock = &klock;
        lockd_req * 		rq;
        nlm_lock *		nlock = &(argp->alock);
        nlm_res			nres;
        int			ex = argp->exclusive;

	if ((klock.fh.len = nlock->fh.n_len) > sizeof(klock.fh.data))
		return 0;
	memcpy(klock.fh.data, nlock->fh.n_bytes, klock.fh.len);
        strncpy(klock.server_name, host->name, sizeof(klock.server_name));
        klock.l_offset = nlock->l_offset;
        klock.l_len = nlock->l_len;
        klock.pid = nlock->svid;

        /* Check if there's a pending lock request that is still waiting
         * for its reply/callback. This may happen when the server's 
         * reply got lost, but the grant notification didn't.
         */
        if ((rq = req_find(host->l_waiting, RQ_LOCK, alock, ex, -1)) != NULL) {
        	goto foundit;
        }

        /* Find the corresponding blocked lock. If we don't find it, return 0
         * to make the caller return an error to the granting NLM server.
         * It is important for the server to know that a grant request failed,
         * otherwise the granted lock won't be owned by anyone.
         */
        if ((rq = req_find(host->l_blocked, RQ_LOCK, alock, ex, -1)) == NULL) {
		log(L_WARNING, "%s granted unknown lock", host->name);
                return 0;
        }

foundit:
        /* Found it. */
        /* req_unlink(rq); This is done in req_process_result */

	/* Update list of local locks */
	nlm_update_local(AC_LOCK, rq);

	/* Propagate the result to the client */
	req_process_result(rq, RQ_GRANTED);

        return 1;
}

/*
 * The server has accepted a lock or unlock request. Update our local
 * list of locks.
 */
static void
nlm_update_local(int action, struct lockd_req *rq)
{
	lockd_lock	lock;
	lockd_file *	file;
	netobj		owner_o;

	if (rq == NULL) return;

	if ((file = file_lookup(rq->host, &(rq->fh), 0)) == NULL) {
		log(L_ERROR, "couldn't create file struct for remote file");
		goto bonkers;
	}

	nlm_create_oh(&owner_o, rq->pid);
	lk_quick_init(&lock, rq->host, file, rq->l_offset, rq->l_len,
				&owner_o, rq->pid, rq->exclusive);
	if (lk_lock(action, &lock) != LCK_GRANTED) {
		log(L_ERROR, "couldn't add remote lock to local list");
		goto bonkers;
	}
	return;

bonkers:
	/* FIXME: should do something useful here. */
}
#endif

/*
 * Initialize request struct, protocol-specific part.
 */
static int
nlm_rqinit_1(lockd_req *rq)
{
	nlm_create_oh(&rq->oh, rq->pid);
	rq->cookie = rpcmsg_gencookie();
	return 1;
}

/*
 * Send an NLM request.
 */
static int
nlm_rqsend_1(lockd_req *rq)
{
	lockd_host *	host;
	struct in_addr	addr;
	int		cookie;
	netobj		owner_o, cookie_o;
	int		status;

	host = rq->host;
	addr = host->addr;

	nlm_create_oh(&owner_o, rq->pid);

	cookie = rq->cookie;
	cookie_o.netobj_data = (void *) &cookie;
	cookie_o.netobj_len  = sizeof(cookie);

	switch (rq->type) {
	case RQ_TEST: {
		nlm_testargs	args;

		netobj_init(&(args.alock.fh), &(rq->fh));
		netobj_init(&(args.alock.oh), &owner_o);
		netobj_init(&(args.cookie), &cookie_o);

		args.exclusive = rq->exclusive;
		args.alock.caller_name = host_myname();
		args.alock.svid = rq->pid;
		args.alock.l_offset = rq->l_offset;
		args.alock.l_len = rq->l_len;

		log(L_DEBUG, "calling NLM_TEST_MSG on %s", inet_ntoa(addr));
		status = rpcmsg_send(addr, nlm_svcpg, NLM_VERS, NLM_TEST_MSG,
				(xdrproc_t)xdr_nlm_testargs, &args);

		break;
		}

	case RQ_LOCK: {
		nlm_lockargs 	args;

		netobj_init(&(args.alock.fh), &(rq->fh));
		netobj_init(&(args.alock.oh), &owner_o);
		netobj_init(&(args.cookie), &cookie_o);

		args.block = rq->block;
		args.exclusive = rq->exclusive;
		args.reclaim = rq->reclaim;
		args.state = 0;	/* we don't do NSM yet */

		args.alock.caller_name = host_myname();
		args.alock.svid = rq->pid;
		args.alock.l_offset = rq->l_offset;
		args.alock.l_len = rq->l_len;

		log(L_DEBUG, "calling NLM_LOCK_MSG on %s", inet_ntoa(addr));
		status = rpcmsg_send(addr, nlm_svcpg, NLM_VERS, NLM_LOCK_MSG,
				(xdrproc_t)xdr_nlm_lockargs, &args);

		break;
		}

	case RQ_CANCEL: {
		nlm_cancargs	args;

		netobj_init(&(args.alock.fh), &(rq->fh));
		netobj_init(&(args.alock.oh), &owner_o);
		netobj_init(&(args.cookie), &cookie_o);

		args.alock.caller_name = host_myname();
		args.alock.svid = rq->pid;
		args.alock.l_offset = rq->l_offset;
		args.alock.l_len = rq->l_len;
		args.exclusive = rq->exclusive;
		args.block = rq->block;

		log(L_DEBUG, "calling NLM_CANCEL_MSG on %s", inet_ntoa(addr));
		status = rpcmsg_send(addr, nlm_svcpg, NLM_VERS, NLM_CANCEL_MSG,
				(xdrproc_t)xdr_nlm_cancargs, &args);

		break;
		}

	case RQ_UNLOCK: {
		nlm_unlockargs	args;

		netobj_init(&(args.alock.fh), &(rq->fh));
		netobj_init(&(args.alock.oh), &owner_o);
		netobj_init(&(args.cookie), &cookie_o);

		args.alock.caller_name = host_myname();
		args.alock.svid = rq->pid;
		args.alock.l_offset = rq->l_offset;
		args.alock.l_len = rq->l_len;

		log(L_DEBUG, "calling NLM_UNLOCK_MSG on %s", inet_ntoa(addr));
		status = rpcmsg_send(addr, nlm_svcpg, NLM_VERS, NLM_UNLOCK_MSG,
				(xdrproc_t)xdr_nlm_unlockargs, &args);

		break;
		}
	}

	return status;
}

static int
nlm_getconflock_1(lockd_req *rq, struct lklm_holder *hop)
{
	struct nlm_testres	*resp;
	struct nlm_holder	*h;

	if (rq != RQ_TEST || !rq->result)
		return 0;
	resp = (struct nlm_testres *) rq->result;
	h = &(resp->stat.nlm_testrply_u.holder);

	strcpy(hop->alock.server_name, "<foobar>");
	hop->alock.fh.len = 0;
	hop->alock.pid = h->svid;
	hop->alock.l_offset = h->l_offset;
	hop->alock.l_len = h->l_len;
	hop->exclusive = h->exclusive;
	return 1;
}

/*
 * Create an owner handle. We don't bother with allocating memory and
 * such because these handles will be used only as an argument to 
 * rpcmsg_send, and are not kept around.
 */
#define MAXCOOKIEHOST	(MAX_NETOBJ_SZ - sizeof(int))
static void
nlm_create_oh(netobj *obj, int pid)
{
	static char	ownerhandle[MAX_NETOBJ_SZ];
	static int	ownerpidoff = 0;

	/* Create the owner handle. This is the concatenation of 
	 * our hostname and the caller's pid.
	 */
	if (ownerpidoff == 0) {
		strncpy(ownerhandle, host_myname(), MAXCOOKIEHOST);
		ownerhandle[MAXCOOKIEHOST] = '\0';
		if ((ownerpidoff = strlen(ownerhandle)) == 0) {
			log(L_ERROR, "can't get my own hostname");
			strcpy(ownerhandle, "foobar");
			ownerpidoff = 6;
		}
	}
	memcpy(ownerhandle + ownerpidoff, &pid, sizeof(int));
	obj->netobj_data = ownerhandle;
	obj->netobj_len  = ownerpidoff+sizeof(int);
}

#if 0
static int
nlm_rqstatus(int nlmstat)
{
static struct {
	int	nlmstat;
	int	rqstat;
} statusmap[] = {
	{ LCK_GRANTED,		RQ_GRANTED },
	{ LCK_DENIED,		RQ_DENIED },
	{ LCK_DENIED_NOLOCKS,	RQ_DENIED_NOLOCKS },
	{ LCK_BLOCKED,		RQ_BLOCKED },
	{ LCK_DENIED_GRACE_PERIOD, RQ_DENIED_GRACE_PERIOD },
	{ -1,			RQ_FAILED },
};
	int	i;

	for (i = 0; statusmap[i].nlmstat >= 0; i++)
		if (statusmap[i].nlmstat == nlmstat)
			break;

	return statusmap[i].rqstat;
}
#endif
